﻿using System;
using System.Net.Sockets;
using System.Threading;

namespace KathreinApi2008LanOnly
{
    public delegate void ReadEpcsEventHandler(byte[] epc);
    public delegate void VoidEventHandler();

    public sealed class Reader : IDisposable
    {
        private static byte[] start = new byte[] { 0xaa, 0xbb };
        private static byte[] end = new byte[] { 0xaa, 0xcc };

        // Command constants
        private static byte[] readEpcsCommand = new byte[] { 0x01, 0x01 };
        private static byte[] readEpcsCommandData = new byte[] { 0x00 };
        private static byte[] writeEpcsCommand = readEpcsCommand;
        private static byte[] writeEpcsCommandDataPrefix = new byte[] { 0x09, 0, 0, 0, 0, 0x60 };

        private TcpClient client;
        private bool exitThread;
        private bool faulted;
        private string hostname;
        private object lockRoot = new object();
        private NetworkStream stream;
        private Thread thread;

        public event ReadEpcsEventHandler ReadEpcs;
        public event VoidEventHandler WriteEpcNoAnswerFromTag;
        public event VoidEventHandler WriteEpcNoTagInField;
        public event VoidEventHandler WriteEpcNotEnoughEnergy;
        public event VoidEventHandler WriteEpcSuccessful;

        public bool Faulted
        {
            get
            {
                lock(lockRoot)
                    return faulted;
            }

            private set
            {
                lock(lockRoot)
                    faulted = value;
            }
        }

        public Reader(string hostname)
        {
            this.hostname = hostname;
        }

        private void BackgroundThread()
        {
            byte[] buffer = new byte[2];

            try
            {
                while(true)
                {
                    lock(lockRoot)
                        if(exitThread)
                            break;

                    if(stream.Read(buffer, 0, 2) != 2)
                        throw new ProtocolException("Start marker not received");

                    if(buffer[0] != 0xAA || buffer[1] != 0xBB)
                        throw new ProtocolException("Start marker invalid");

                    if(stream.Read(buffer, 0, 2) != 2)
                        throw new ProtocolException("Command not received");

                    ReadDataForCommandAndCallEvents(buffer);

                    if(stream.Read(buffer, 0, 2) != 2)
                        throw new ProtocolException("End marker not received");

                    if(buffer[0] != 0xAA || buffer[1] != 0xCC)
                        throw new ProtocolException("End marker invalid");
                }
            }
            catch
            {
                Faulted = true;

                // Put stream in closed state on faults
                stream.Close();
                throw;
            }
        }

        public void Connect()
        {
            Disconnect();

            client = new TcpClient(hostname, 4007);
            client.NoDelay = true;

            stream = client.GetStream();

            // This is sent first by ReaderStart, so just send it as well
            stream.Write(new byte[] { 0xaa, 0xbb, 0x01, 0x01, 0x44, 0xaa, 0xcc }, 0, 7);

            var buffer = new byte[13];
            if(stream.Read(buffer, 0, 13) != 13)
                throw new ProtocolException("Response to initial packet was incorrect");

            StartBackgroundThread();
        }

        public void Disconnect()
        {
            if(client != null)
            {
                StopBackgroundThread();

                stream.Close();
                client.Close();
                client = null;
            }

            Faulted = false;
        }

        public void Dispose()
        {
            if(client != null)
                Disconnect();
        }

        public void ReadEpcsAsync()
        {
            SendCommand(readEpcsCommand, readEpcsCommandData);
        }

        private void ReadDataForCommandAndCallEvents(byte[] buffer)
        {
            if(buffer[0] == 0x07 && buffer[1] == 0x00)
            {
                // No data
            }
            else if(buffer[0] == 0x01 && buffer[1] == 0x01)
            {
                // Response to ReadEpcs

                var nextBytes = new byte[3];

                if(stream.Read(nextBytes, 0, 3) != 3)
                    throw new ProtocolException("Unexpected end of stream");

                if(nextBytes[0] == 0x80)
                {
                    if(nextBytes[1] == 0 && nextBytes[2] == 0)
                    {
                        // Nothing else to do
                    }
                    else if(nextBytes[1] == 1)
                    {
                        // EPCs from antenna 1

                        var epc = ReadEpc();

                        if(ReadEpcs != null)
                            ReadEpcs(epc);
                    }
                    else
                        throw new ProtocolException("ReadEpcs response not understood");
                }
                else if(nextBytes[0] == 0x85)
                {
                    if(nextBytes[1] == 0 && nextBytes[2] == 0)
                    {
                        if(WriteEpcSuccessful != null)
                            WriteEpcSuccessful();
                    }
                    else if(nextBytes[1] == 2 && nextBytes[2] == 0)
                    {
                        if(WriteEpcNoTagInField != null)
                            WriteEpcNoTagInField();
                    }
                    else if(nextBytes[1] == 4 && nextBytes[2] == 0)
                    {
                        if(WriteEpcNoAnswerFromTag != null)
                            WriteEpcNoAnswerFromTag();
                    }
                    else if(nextBytes[1] == 0x0B && nextBytes[2] == 0x01)
                    {
                        if(WriteEpcNotEnoughEnergy != null)
                            WriteEpcNotEnoughEnergy();
                    }
                    else
                        throw new ProtocolException("ReadEpcs response not understood");
                }
                else
                    throw new ProtocolException("ReadEpcs response not understood");
            }
            else
                throw new ProtocolException("Command not supported, don't know how much data to read");
        }

        private byte[] ReadEpc()
        {
            var epcLittleEndian = new byte[12];
            if(stream.Read(epcLittleEndian, 0, 12) != 12)
                throw new ProtocolException("Unexpected end of stream when reading EPC");

            var epc = new byte[12];
            for(int i = 0; i < 12; ++i)
                epc[i] = epcLittleEndian[11 - i];

            return epc;
        }

        private void SendCommand(byte[] command, byte[] data)
        {
            if(Faulted)
                throw new InvalidOperationException("Reader class instance is in faulted state, you have to connect again");

            if(command.Length != 2)
                throw new ArgumentException("Command must be two bytes long");

            stream.Write(start, 0, 2);
            stream.Write(command, 0, 2);
            stream.Write(data, 0, data.Length);
            stream.Write(end, 0, 2);
        }

        private void StartBackgroundThread()
        {
            if(thread != null)
                throw new InvalidOperationException("Background thread already started");

            exitThread = false;

            thread = new Thread(BackgroundThread);
            thread.IsBackground = true;
            thread.Start();
        }

        private void StopBackgroundThread()
        {
            if(thread == null)
                throw new InvalidOperationException("Background thread not started");

            lock(lockRoot)
                exitThread = true;

            if(!thread.Join(500))
                thread.Abort();
        }

        public void WriteEpcAsync(byte[] epc)
        {
            if(epc.Length != 12)
                throw new ArgumentException("EPC must be 12 bytes (96 bits) long");

            var data = new byte[writeEpcsCommandDataPrefix.Length + 12];
            writeEpcsCommandDataPrefix.CopyTo(data, 0);

            for(int i = 0; i < 12; ++i)
                data[writeEpcsCommandDataPrefix.Length + i] = epc[11 - i];

            SendCommand(writeEpcsCommand, data);
        }
    }
}
